function [finished ] = learn_kalman(data_ca, ct_ca, affil_ca)
% LEARN_KALMAN Find the ML parameters of a stochastic Linear Dynamical System using EM.
%
% [A, C, Q, R, INITX, INITV, LL] = LEARN_KALMAN(DATA, A0, C0, Q0, R0, INITX0, INITV0) fits
% the parameters which are defined as follows
%   x(t+1) = A*x(t) + w(t),  w ~ N(0, Q),  x(0) ~ N(init_x, init_V)
%   y(t)   = C*x(t) + v(t),  v ~ N(0, R)
% A0 is the initial value, A is the final value, etc.
% DATA(:,t,l) is the observation vector at time t for sequence l. If the sequences are of
% different lengths, you can pass in a cell array, so DATA{l} is an O*T matrix.
% LL is the "learning curve": a vector of the log lik. values at each iteration.
% LL might go positive, since prob. densities can exceed 1, although this probably
% indicates that something has gone wrong e.g., a variance has collapsed to 0.
%
% There are several optional arguments, that should be passed in the following order.
% LEARN_KALMAN(DATA, A0, C0, Q0, R0, INITX0, INITV0, MAX_ITER, DIAGQ, DIAGR, ARmode)
% MAX_ITER specifies the maximum number of EM iterations (default 10).
% DIAGQ=1 specifies that the Q matrix should be diagonal. (Default 0).
% DIAGR=1 specifies that the R matrix should also be diagonal. (Default 0).
% ARMODE=1 specifies that C=I, R=0. i.e., a Gauss-Markov process. (Default 0).
% This problem has a global MLE. Hence the initial parameter values are not important.
% 
% LEARN_KALMAN(DATA, A0, C0, Q0, R0, INITX0, INITV0, MAX_ITER, DIAGQ, DIAGR, F, P1, P2, ...)
% calls [A,C,Q,R,initx,initV] = f(A,C,Q,R,initx,initV,P1,P2,...) after every M step. f can be
% used to enforce any constraints on the params. 
%
% For details, see
% - Ghahramani and Hinton, "Parameter Estimation for LDS", U. Toronto tech. report, 1996
% - Digalakis, Rohlicek and Ostendorf, "ML Estimation of a stochastic linear system with the EM
%      algorithm and its application to speech recognition",
%       IEEE Trans. Speech and Audio Proc., 1(4):431--442, 1993.


%    learn_kalman(data, A, C, Q, R, initx, initV, max_iter, diagQ, diagR, ARmode, constr_fun, varargin)



% convert the cell arrays from java to matrix 

nrow = length(data_ca);
data_all = [];
affil_all = [];
for i = 1:nrow
	data_all = [data_all ; cell2mat(data_ca{i})];
    affil_all = [affil_all ; cell2mat(affil_ca{i})];
end
ct = cell2mat(ct_ca{1});



if length(find(data_all > 100)) > 0
    fprintf('need to impute gene expression levels');
end


% first select a subset of genes which have GO annotations

annotatedgenes = find(sum(affil_all,2) > 0);
cutoff = 0.1;
while length(annotatedgenes) > 1000
    temp = find(max(data_all(annotatedgenes, :),[],2) - min(data_all(annotatedgenes, :),[],2) > cutoff);
    annotatedgenes = annotatedgenes(temp);
    cutoff  =  cutoff + 0.1;
end
data = data_all(annotatedgenes,:);
affil = affil_all(annotatedgenes,:);

fprintf('%d genes selected for inferring the underlying processes.\n' , size(data,1));


finished = '0';
N = length(ct)-1;
ss = size(affil,2);
os = size(affil,1);
A = eye(ss);
C = affil;
Q = zeros(ss,ss,N);
initx = zeros(ss,N);
for d = 1:N
    Q(:,:,d) = 0.01*eye(ss);
    R(d) = 0.01;
    B(d) = 0.015;
    initV(:,:,d) = eye(ss);
end
verbose = 1;
max_iter = 2;
thresh = 1e-4;
diagQ = 1;
diagR = 1;
ARmode = 0;
lambda = 0.5;
help = 10 * ones(N,1);
option =  optimset('Display','off','LargeScale','off');
selectedprocess = zeros(N,ss);
for j = 1:ss
        selectedprocess(:,j) = j;
end
full = 0;





alpha = zeros(os, os, N);
Tsum = zeros(N,1);
for d = 1:N
  %y = data(:,:,ex);
  y = data(:,ct(d)+1:ct(d+1));
  T = size(y,2);
  Tsum(d) = T;
  alpha_temp = zeros(os, os);
  for t=1:T
    alpha_temp = alpha_temp + y(:,t)*y(:,t)';
  end
  alpha(:,:,d) = alpha(:,:,d) + alpha_temp;
end

previous_loglik = -inf;
loglik = 0;
converged = 0;
num_iter = 1;
LL = [];

numprocess = size(selectedprocess,2);
% Convert to inline function as needed.
%if ~isempty(constr_fun)
%  constr_fun = fcnchk(constr_fun,length(varargin));
%end

noedge = find(sum(C,2) == 0);

while ~converged & (num_iter <= max_iter) 

  %%% E step
  
  delta = zeros(os, ss,N);
  gamma = zeros(ss, ss,N);
  gamma1 = zeros(ss, ss,N);
  gamma2 = zeros(ss, ss,N);
  beta = zeros(ss, ss,N);
  P1sum = zeros(ss, ss,N);
  x1sum = zeros(ss, 1,N);
  loglik = 0;
  xsmooth = zeros(ss,ct(length(ct)));
  %C = sparse(C);
  
%  for ex = 1:N
%    y = data{ex};
    for d = 1:N
        y = data(:,ct(d)+1:ct(d+1));
        Qt = diag(Q(:,:,d));
        Qt = Qt(selectedprocess(d,:));
        Qt = diag(Qt);
        Vt = initV(selectedprocess(d,:),selectedprocess(d,:),d);
        Xt = initx(selectedprocess(d,:),d);
        tempR = repmat(R(d),os,1);
        tempR(noedge) = B(d);
        tempR = diag(tempR);
        [beta_t, gamma_t, delta_t, gamma1_t, gamma2_t, x1, V1, loglik_t, xsmooth(selectedprocess(d,:),ct(d)+1:ct(d+1)),help(d)] = ...
        Estep(y, eye(numprocess), C(:,selectedprocess(d,:)) , Qt, tempR, Xt, Vt, ARmode, num_iter,help(d));
        
  %      result = sortrows([ sum(xsmooth(:,ct(d)+1:ct(d+1)).^2,2)  (1:108)'],1);
  %      xsmooth(result(1:88,2),ct(d)+1:ct(d+1)) = 0 ; 
        
        beta(selectedprocess(d,:),selectedprocess(d,:),d) = beta_t;
 %       gamma = gamma + gamma_t;
        delta(:,selectedprocess(d,:),d) = delta_t;
        gamma1(selectedprocess(d,:),selectedprocess(d,:),d) = gamma1_t;
        gamma2(selectedprocess(d,:),selectedprocess(d,:),d) = gamma2_t;
        P1sum(selectedprocess(d,:),selectedprocess(d,:),d) = V1 + x1*x1';
        x1sum(selectedprocess(d,:),d) = x1;
        %fprintf(1, 'example %d, ll/T %5.3f\n', ex, loglik_t/T);
        loglik = loglik + loglik_t;
    end
%  end
  %LL = [LL loglik];

  LL = loglik;
  num_iter =  num_iter + 1;

  %%% M step
  
  % Tsum =  N*T
  % Tsum1 = N*(T-1);
  Tsum1 = sum(Tsum) - N;
  % %%% A = beta * inv(gamma1);
  %A = (gamma1' \ beta')';
%  for d = 1:N
%      Q(:,:,d) = (gamma2(:,:,d) - A*beta(:,:,d)') / Tsum(d);
%      if diagQ
%        Q(:,:,d) = diag(diag(Q(:,:,d)));
%      end
%  end
  if ~ARmode
    
%    C = zeros(os,ss);
%    temp = find(sum(affil,2) ~= 0 );
%    for i = 1:length(temp)
%        cur = find(affil(temp(i),:) == 1);        
%        X = xsmooth(cur,:)';
%        Y = y(temp(i),:)';
%        b = inv(X' * X) * X' * Y;
%        C(temp(i),cur) = b;
%   end

%    C = zeros(os,ss);
    
%    if full == 0
%        fprintf('calculate gene ');
%        for i = 1:100  %os
%            w = fmincon(@func, C(i,:)' ,-1*eye(ss),zeros(ss,1),[],[],[],[],[],option, xsmooth, data{1}(i,:)', lambda , affil(i,:)', R , N,ct);
%            C(i,:) = w';
%            fprintf('%d ',i);
%            if rem(i,30) == 0
%                fprintf('\n');
%            end
%        end
%        fprintf('\n');
%    end
 
    noedge = [];
    if full == 0
 %       fprintf('calculate gene ');
        for i = 1:os
            

            value0 = 0;
            for d = 1: N
                value0 = value0 + sum( (data(i,ct(d)+1:ct(d+1)).^2)' / B(d))   ;
            end

            
            for p  = 1:ss
                [w1(p), eval(p)] = fmincon(@func, C(i,p)' ,-1,0,[],[],[],[],[],option, xsmooth(p,:), data(i,:)', lambda , affil(i,p), R , N,ct);
            end
            
            [value1 , index1] = min(eval);
            w2 = zeros(2,ss);
            for p = 1:ss
                if p == index1 
                    eval(p) = Inf;
                    continue;
                end
                [w2(1:2,p), eval(p)] = fmincon(@func, C(i,[index1 p])' ,-1 * eye(2),[0;0],[],[],[],[],[],option, xsmooth([index1 p],:), data(i,:)', lambda , affil(i,[index1 p])', R , N,ct);
            end
            [value2 , index2] = min(eval);

            C(i,:) = 0;   
            
            [junk, in] = min([value0 value1 value2]);
            
            if in == 1
                noedge = [noedge i];
            elseif in == 2
               C(i, index1) = w1(index1);
            elseif in == 3
                C(i,[index1 index2]) = w2(1:2,index2)';
            end
            
%            fprintf('%d ',i);
%            if rem(i,30) == 0
%                fprintf('\n');
%            end

        end
        fprintf('\n');
    end
 
 

    %C = sparse(C);
    % yanxin   C = delta * inv(gamma);
    %C = (gamma' \ delta')';
    
    initx = x1sum / 1;
    tempR = zeros(os,os,N);
  	for d = 1:N
        initV(:,:,d) = P1sum(:,:,d) / 1 - initx(:,d)*initx(:,d)';
        initV(:,:,d) = diag(diag(initV(:,:,d)));
        tempR(:,:,d) = (alpha(:,:,d) - C*delta(:,:,d)') / Tsum(d);
        if diagR
            wholegene = 1:os;
            foredgedgene = setdiff(wholegene,noedge);
            R(d) = mean(diag(tempR(foredgedgene,foredgedgene,d)));
            if length(noedge) ~=0
                B(d) = mean(diag(tempR(noedge,noedge,d)));
            end
        end
    end
  end


%   if ~isempty(constr_fun)
%     [A,C,Q,R,initx,initV] = feval(constr_fun, A, C, Q, R, initx, initV, varargin{:});
%   end
  
 % converged = em_converged(loglik, previous_loglik, thresh);
  previous_loglik = loglik;
end



% then calculate the structure for the whole model

weight_all = zeros(size(affil_all,1),size(affil_all,2));
weight_all(annotatedgenes,:) = C;

unannotatedgenes = setdiff(1:size(affil_all,1), annotatedgenes);
%fprintf('calculating new gene ');

for i = 1:length(unannotatedgenes)
    curgene = unannotatedgenes(i);
    

    value0 = 0;
    for d = 1: N
        value0 = value0 + sum( (data_all(curgene,ct(d)+1:ct(d+1)).^2)' / B(d))   ;
    end


    for p  = 1:ss
        [w1(p), eval(p)] = fmincon(@func, weight_all(curgene,p)' ,-1,0,[],[],[],[],[],option, xsmooth(p,:), data_all(curgene,:)', lambda , affil_all(curgene,p), R , N,ct);
    end

    [value1 , index1] = min(eval);
    w2 = zeros(2,ss);
    for p = 1:ss
        if p == index1 
            eval(p) = Inf;
            continue;
        end
        [w2(1:2,p), eval(p)] = fmincon(@func, weight_all(curgene,[index1 p])' ,-1 * eye(2),[0;0],[],[],[],[],[],option, xsmooth([index1 p],:), data_all(curgene,:)', lambda , affil_all(curgene,[index1 p])', R , N,ct);
    end
    [value2 , index2] = min(eval);

    weight_all(curgene,:) = 0;   

    [junk, in] = min([value0 value1 value2]);

    if in == 2
       weight_all(curgene, index1) = w1(index1);
    elseif in == 3
        weight_all(curgene,[index1 index2]) = w2(1:2,index2)';
    end

%    fprintf('%d ',i);
%    if rem(i,30) == 0
 %       fprintf('\n');
 %   end

end

fprintf('\n');

dlmwrite('xsmooth',xsmooth,'\t');
dlmwrite('weight',weight_all,'\t');

finished = '1';


%%%%%%%%%

function [beta, gamma, delta, gamma1, gamma2, x1, V1, loglik, xsmooth,help] = ...
    Estep(y, A, C, Q, R, initx, initV, ARmode , num_iter,help)
%
% Compute the (expected) sufficient statistics for a single Kalman filter sequence.
%

[os T] = size(y);
ss = length(A);

if ARmode
  xsmooth = y;
  Vsmooth = zeros(ss, ss, T); % no uncertainty about the hidden states
  VVsmooth = zeros(ss, ss, T);
  loglik = 0;
else
  [xsmooth, Vsmooth, VVsmooth, loglik,help] = kalman_smoother(y, A, C, Q, R, initx, initV, num_iter,help);
end

delta = zeros(os, ss);
gamma = zeros(ss, ss);
beta = zeros(ss, ss);
for t=1:T
  delta = delta + y(:,t)*xsmooth(:,t)';
  gamma = gamma + xsmooth(:,t)*xsmooth(:,t)' + Vsmooth(:,:,t);
  if t>1 beta = beta + xsmooth(:,t)*xsmooth(:,t-1)' + VVsmooth(:,:,t); end
end
gamma1 = gamma - xsmooth(:,T)*xsmooth(:,T)' - Vsmooth(:,:,T);
gamma2 = gamma - xsmooth(:,1)*xsmooth(:,1)' - Vsmooth(:,:,1);

x1 = xsmooth(:,1);
V1 = Vsmooth(:,:,1);

